#!/usr/bin/env python3
import traceback
from datetime import datetime, timezone
import os

import xlsxwriter
from xlsxwriter.utility import xl_col_to_name

from .base import ReportGenerator


class XlsxReportGenerator(ReportGenerator):
    """
    Report generator that produces an Excel report for a single agent.
    The report consists of two sheets:
      - A "Data" sheet that contains a header block and the raw data.
      - A "Summary" sheet that has the same header block and a chart that references the Data sheet.

    The class is structured so that each section is implemented in a separate method,
    making it easier for subclasses to override or extend parts of the report.
    """

    include_summary_sheet = True

    cursor_cache_values = True

    logo_filename = "logo.png"
    logo_cell = "B2"
    header_logo_x_offset = 1
    header_logo_y_offset = 5

    chart_scale = {"x_scale": 2, "y_scale": 2}
    include_x_axis = False
    series_kwargs = {}  ## Styling applied to all series
    filtered_series_kwargs = {}  ## Styling applied to specific series
    summary_chart_height = 30
    exclude_chart_series = []

    legend = {"position": "bottom"}
    y1_axis = {"major_gridlines": {"visible": False}}
    y2_axis = {}
    y2_axis_series = []

    def generate_one(self, agent_id: str) -> None:
        """Orchestrates the creation of the Data and Summary sheets for one agent."""
        agent_display_name = self.get_agent_display_name(agent_id=agent_id)
        self.add_to_log(f"Generating report for {agent_display_name}...")

        series, msgs = self.retrieve_report_data(agent_id)
        self.parse_data_series(series, msgs)

        self.workbook = self.open_workbook(workbook_name=agent_display_name)

        ## Create the summary sheet first to put it first, then put aside for later
        if self.include_summary_sheet:
            summary_sheet = self.create_sheet("Summary")

        # Create and fill the Data sheet.
        data_sheet = self.create_sheet("Data")
        data_range = self.write_data_sheet(data_sheet)

        # Fill the Summary sheet.
        if self.include_summary_sheet:
            self.write_summary_sheet(summary_sheet, data_range)

        self.close_workbook()

    # -- Setup and Data Retrieval --

    def retrieve_report_data(self, agent_id: str):
        """Retrieve the available series and data messages."""
        self.add_to_log("Finding available dataseries...")
        agg = self.get_current_data_aggregate(agent_id)
        series = self.get_available_dataseries(agg)
        self.add_to_log(f"Available series: {series}")
        self.update_progress(0.05)

        self.add_to_log("Retrieving data...")
        msgs = self.retrieve_data(
            period_from=self.localized_period_from,
            period_to=self.localized_period_to,
            agent_id=agent_id,
        )
        self.add_to_log(f"Finished retrieving {len(msgs)} messages")
        self.update_progress(0.20)
        return series, msgs

    # -- Workbook and Common Header --

    def open_workbook(self, workbook_name=None):
        """Create and return a new workbook."""
        workbook_name = self.get_ws_path(f"{workbook_name}.xlsx")
        wb = xlsxwriter.Workbook(workbook_name)
        self.wb_format_title = wb.add_format({"bold": True, "font_size": 14})
        self.wb_format_subtitle = wb.add_format({"italic": True, "font_size": 8})
        self.wb_format_bold = wb.add_format({"bold": True})
        self.wb_format_datetime_format = wb.add_format(
            {"num_format": "dd/mm/yyyy hh:mm:ss"}
        )
        self.wb_format_date_format = wb.add_format({"num_format": "dd/mm/yy"})
        self.wb_format_short_date_format = wb.add_format({"num_format": "dd/mm"})
        return wb

    def close_workbook(self):
        """Close the workbook."""
        self.workbook.close()

    def write_header_block(self, sheet) -> tuple[int, int]:
        """
        Write the common header block into the given sheet.
        Returns a tuple: (next_row, index_col) where next_row is the first row after the header.
        """
        index_col = 1
        next_row = 1

        if self.is_logo_available():
            index_col = self.header_logo_x_offset
            sheet.insert_image(
                self.logo_cell, self.logo_filename, {"x_scale": 0.5, "y_scale": 0.5}
            )
            next_row = self.header_logo_y_offset

        sheet.write(next_row, index_col, self.report_name, self.get_title_format())
        next_row += 1
        sheet.write(
            next_row, index_col, "Auto-generated by Doover", self.get_subtitle_format()
        )
        next_row += 2

        # Write time period details.
        sheet.write(next_row, index_col, "Start : ", self.get_bold_format())
        sheet.write(
            next_row,
            index_col + 2,
            self.localized_period_from.strftime("%d/%m/%y %H:%M:%S"),
        )
        next_row += 1
        sheet.write(next_row, index_col, "End : ", self.get_bold_format())
        sheet.write(
            next_row,
            index_col + 2,
            self.localized_period_to.strftime("%d/%m/%y %H:%M:%S"),
        )

        next_row += 2
        # sheet.set_column(index_col, index_col, 20)
        sheet.write(next_row, index_col, "Timezone : ", self.get_bold_format())
        sheet.write(next_row, index_col + 2, str(self.for_timezone))

        next_row += 3

        return next_row, index_col

    def is_logo_available(self):
        """Check if the logo file is available."""
        if self.logo_filename is None:
            return False
        if os.path.exists(self.logo_filename):
            return True
        if self.logging_function:
            self.logging_function("Logo file not accessible")
        return False

    def create_sheet(self, sheet_name="Data"):
        """Create the 'Data' sheet"""
        sheet = self.workbook.add_worksheet(sheet_name)
        return sheet

    # -- Data Sheet Methods --

    def process_message(self, m, series, value_cache):
        """
        Process a single message `m` and return a tuple:
        (temp_values, local_dt_str)

        - temp_values: List of computed values for each series. if temp_values[i] is None, the value is skipped.
        - local_dt_str: The formatted timestamp string.

        Subclasses can override this method to change how individual messages are processed.
        """
        temp_values = {}

        for i, s in enumerate(series):
            try:
                temp_obj = self.get_data_object(m["payload"], s["name"])
                value, raw_value = None, None
                if (
                    temp_obj
                    and "currentValue" in temp_obj
                    and temp_obj["currentValue"] is not None
                ):
                    current_val = temp_obj["currentValue"]
                    if isinstance(current_val, float):
                        current_val = round(current_val, 3)
                    value = current_val
                    raw_value = current_val
                elif self.cursor_cache_values and s["name"] in value_cache:
                    value = value_cache[s["name"]]

                if value is not None:
                    # Example special condition.
                    if s["name"] == "lastPacket" and raw_value in [
                        "Flow",
                        "Calibration",
                    ]:
                        return {}
                    temp_values[s["name"]] = value
                    value_cache[s["name"]] = value
            except Exception:
                if self.test_mode:
                    self.add_to_log(
                        f"Error getting datapoint {s['name']} : {traceback.format_exc()}"
                    )

        return temp_values

    def parse_data_series(self, series, msgs):
        """
        Process the incoming messages (msgs) and build the self.data_series list.
        Each element in self.data_series is a dictionary with keys:
        - "name": the series' internal name.
        - "display_name": the series' display string.
        - "timestamp_strs": list of X values (as formatted strings).
        - "timestamps": list of X values (as datetime objects).
        - "values": list of Y values.
        """
        value_cache = {}
        # Initialize self.data_series for each series.
        self.data_series = []
        for s in series:
            self.data_series.append(
                {
                    "name": s["name"],
                    "display_name": s["displayString"],
                    "timestamp_strs": [],
                    "timestamps": [],
                    "values": [],
                }
            )

        # Process each message using the helper function.
        for m in msgs:
            temp_values = self.process_message(m, series, value_cache)
            local_dt = datetime.fromtimestamp(m["timestamp"], timezone.utc).astimezone()
            local_dt_str = local_dt.strftime("%d/%m/%y %H:%M:%S")

            if len(temp_values.keys()) < 1:
                continue

            # Append the timestamp and values to each series.
            for i, ds in enumerate(self.data_series):
                if ds["name"] in temp_values:
                    ds["timestamp_strs"].append(local_dt_str)
                    ds["timestamps"].append(local_dt)
                    ds["values"].append(temp_values[ds["name"]])

    def get_data_series(self, series_name):
        """Return the data series with the given name."""
        for s in self.data_series:
            if s["name"] == series_name:
                return s
        return None

    def get_data_start_ts(self):
        """Return the start timestamp of the data series."""
        if not self.data_series:
            return None
        return self.data_series[0]["timestamp_strs"][0]

    def get_data_end_ts(self):
        """Return the end timestamp of the data series."""
        if not self.data_series:
            return None
        return self.data_series[0]["timestamp_strs"][-1]

    def write_data_sheet(self, sheet):
        """
        Write the data header and data rows into the 'Data' sheet.
        Assumes that self.data_series has already been populated by parse_data_series.
        Returns a dictionary containing data range information used for charting.
        """
        header_row, _ = self.write_header_block(sheet)

        index_col = 0
        # Write column headers: first the "Time" column, then each series display name.
        sheet.write(header_row, index_col, "Time", self.get_bold_format())
        col_counter = 1
        for s in self.data_series:
            sheet.write(
                header_row,
                index_col + col_counter,
                s["display_name"],
                self.get_bold_format(),
            )
            col_counter += 1

        data_start_row = header_row + 2
        # Determine the number of rows (assuming all series share the same timestamps).
        num_rows = (
            len(self.data_series[0]["timestamp_strs"])
            if self.data_series and self.data_series[0]["timestamp_strs"]
            else 0
        )

        # Write each row of data.
        for i in range(num_rows):
            # Write the timestamp from the first series.
            sheet.write(
                data_start_row + i,
                index_col,
                self.data_series[0]["timestamp_strs"][i],
                self.get_datetime_format(),
            )
            # Write each series' value.
            for j in range(len(self.data_series)):
                value = self.data_series[j]["values"][i]
                if value is not None:
                    sheet.write(data_start_row + i, index_col + j + 1, value)

        self.add_to_log("Finished composing Data sheet")
        self.update_progress(0.80)

        # Return data range information for chart creation.
        data_range = {
            "header_row": header_row,
            "data_start_row": data_start_row,
            "data_end_row": data_start_row + num_rows - 1,
            "data_columns": 1
            + len(self.data_series),  # Column 0 for Time plus one per series.
            "series_names": [s["name"] for s in self.data_series],
        }
        return data_range

    # -- Summary Sheet Methods --

    def write_summary_sheet(self, sheet, data_range):
        """
        Create a chart in the Summary sheet that references the data in the Data sheet.
        The chart will use the data range information from the Data sheet.
        """
        header_row, _ = self.write_header_block(sheet)

        chart = self.get_summary_chart(data_range)

        # Insert the chart into the Summary sheet at a designated position.
        sheet.insert_chart(f"B{header_row + 1}", chart, self.chart_scale)
        self.add_to_log("Inserted chart into Summary sheet")

        self.write_extra_summary(
            sheet, data_range, next_row=header_row + self.summary_chart_height
        )

    def get_summary_chart(self, data_range):
        chart = self.workbook.add_chart({"type": "line"})
        series = self.get_summary_chart_series(data_range)
        for s in series:
            y2_axis_enable = s["variableName"] in self.y2_axis_series
            chart.add_series({**s, "y2_axis": y2_axis_enable})
        # chart.set_title({'name': 'Summary Chart'})
        if self.include_x_axis:
            chart.set_x_axis(
                {
                    "name": "Time",
                    "date_axis": True,
                    # 'num_format': 'dd/mm/yyyy hh:mm:ss'
                    "num_format": "dd/mm",
                }
            )
        else:
            chart.set_x_axis(
                {
                    "line": {"none": True},
                    "label_position": "none",
                    "major_tick_mark": "none",
                    "minor_tick_mark": "none",
                }
            )

        chart.set_y_axis(self.y1_axis)
        if self.y2_axis:
            chart.set_y2_axis(self.y2_axis)

        chart.set_legend(self.legend)

        return chart

    def get_exclude_chart_series(self):
        if hasattr(self, "exclude_chart_series") and self.exclude_chart_series:
            return self.exclude_chart_series
        return []

    def get_summary_chart_series(self, data_range):
        # Chart series for each data column (skip column 0 which is Time).
        series = []
        for col in range(1, data_range["data_columns"]):
            series_name = data_range["series_names"][col - 1]
            if series_name in self.get_exclude_chart_series():
                continue

            # Create the series name from the header row.
            series_name_ref = self.get_series_name_reference(
                "Data", col, data_range["header_row"]
            )

            # Construct the cell range for the categories (x-axis) and values (y-axis).
            categories = self.get_data_sheet_ts_reference(data_range)
            values = self.get_data_sheet_range_reference(data_range, col)

            filtered_kwargs = self.filtered_series_kwargs.get(series_name, {})

            series.append(
                {
                    "variableName": series_name,
                    "name": series_name_ref,
                    "categories": categories,
                    "values": values,
                    **self.series_kwargs,
                    **filtered_kwargs,
                }
            )
        return series

    def write_extra_summary(self, sheet, data_range, next_row):
        """
        Write additional information into the Summary sheet.
        This method can be overridden by subclasses to add more content.
        """
        pass

    # -- Format Helper Methods --

    def get_title_format(self):
        return self.wb_format_title

    def get_subtitle_format(self):
        return self.wb_format_subtitle

    def get_bold_format(self):
        return self.wb_format_bold

    def get_datetime_format(self):
        return self.wb_format_datetime_format

    def get_date_format(self):
        return self.wb_format_date_format

    def get_short_date_format(self):
        return self.wb_format_short_date_format

    def xcol_to_name(self, col):
        return xl_col_to_name(col)

    def to_excel_date(dt):
        # Excel's date system for Windows starts on 1899-12-30.
        excel_start = datetime.datetime(1899, 12, 30)
        delta = dt - excel_start
        return delta.days + (delta.seconds / 86400.0)

    def get_data_sheet_range_reference(self, data_range, col):
        # Get the chart row indices (Excel rows are 1-indexed).
        cat_start, cat_end = self.get_chart_row_indices(data_range)
        # Construct the cell range for the categories (x-axis) and values (y-axis).
        ref = self.get_range_reference("Data", col, cat_start, cat_end)
        return ref

    def get_data_sheet_ts_reference(self, data_range):
        return self.get_data_sheet_range_reference(data_range, 0)

    def get_chart_row_indices(self, data_range):
        """
        Returns the starting and ending row indices for the chart.
        Excel rows are 1-indexed.
        """
        cat_start = data_range["data_start_row"] + 1
        cat_end = data_range["data_end_row"] + 1
        return cat_start, cat_end

    def get_series_name_reference(self, sheet, col, header_row):
        """
        Returns the series name as a cell reference from the header row of the given sheet.
        """
        return f"='{sheet}'!${self.xcol_to_name(col)}{header_row + 1}"

    def get_range_reference(self, sheet, col, row_start, row_end):
        """
        Returns a cell range reference for the given column between row_start and row_end on the specified sheet.
        """
        return f"='{sheet}'!${self.xcol_to_name(col)}{row_start}:${self.xcol_to_name(col)}{row_end}"


generator = XlsxReportGenerator



import asyncio
import logging
import time
import json

from collections.abc import Iterable, Coroutine
from typing import Callable

import grpc

from .grpc_stubs import platform_iface_pb2, platform_iface_pb2_grpc
from .types import Location, Event
from ..grpc_interface import GRPCInterface
from ...utils import maybe_async, call_maybe_async, deprecated
from ...cli.decorators import command as cli_command

log = logging.getLogger(__name__)
PulseCounterCallback = (
    Callable[[int, bool, int, int, str], None]
    | Coroutine[[int, bool, int, int, str], None]
)


class PulseCounter:
    """PulseCounter object for counting pulses on a digital input pin.

    This is generally not created manually, but rather through the PlatformInterface's `get_new_pulse_counter` method.

    Attributes
    ----------
    platform_iface : PlatformInterface
        The platform interface to use for communication.
    pin : int
        The digital input pin number to listen for pulses on.
    edge : str
        The edge to listen for pulses on. Can be "rising", "falling", or "both".
    callback : Callable[[int, bool, int, int, str], Any]
        A callback function to call when a pulse is received. The function should take the following arguments:
            - pin: int - The pin number the pulse was received on.
            - di_value: bool - The value of the digital input pin (1 for high, 0 for low).
            - dt_secs: int - The time since the last pulse in seconds.
            - count: int - The total number of pulses received.
            - edge: str - The edge that the pulse was received on ("rising" or "falling").
        This function can be synchronous or asynchronous.
    rate_window_secs : int
        The size of the window in seconds for which the rate of pulses is calculated.
    count : int
        The total number of pulses received.
    pulse_timestamps : list[float]
        A list of timestamps for each pulse received.
    receiving_pulses : bool
        Whether the PulseCounter is currently receiving pulses.

    """

    def __init__(
        self,
        plt_iface: "PlatformInterface",
        pin: int,
        edge: str = "rising",
        callback: PulseCounterCallback = None,
        rate_window_secs: int = 60,
        auto_start: bool = True,
    ):
        self.platform_iface = plt_iface
        self.pin = pin
        self.edge = edge
        self.callback = callback
        self.rate_window_secs = rate_window_secs
        self.count = 0

        self.start_time = time.time()
        self.pulse_grace_period = (
            0.2  # Need to ignore pulses for a short period after starting
        )
        self.pulse_timestamps = []

        self.receiving_pulses = False
        self.receiving_events = False

        if auto_start:
            self.start_listener_pulses()

    @deprecated("Replaced with PulseCounter.start_listener_pulses")
    def start_listener(self):
        self.start_listener_pulses()

    def start_listener_pulses(self):
        """Start listening for pulses on the digital input pin."""
        # Check each pulse counter is only used for a single type of pulse
        if self.receiving_events:
            log.error("Using a pulse counter for both pulses and offline events")
            return
        self.receiving_pulses = True

        self.start_time = time.time()
        self.platform_iface.start_di_pulse_listener(
            self.pin, self.receive_pulse, edge=self.edge, start_count=self.count
        )

    def update_events(self):
        """Listen for offline events for a digital input pin.

        This is **not** compatible with a PulseCounter that is already receiving pulses.
        """
        # Check each pulse counter is only used for a single type of pulse
        if self.receiving_pulses:
            log.error("Using a pulse counter for both pulses and offline events")
            return
        self.receiving_events = True
        self.receive_events(
            self.platform_iface.get_di_events(
                self.pin, self.edge, include_system_events=True
            )
        )

    def add_existing_events(self, time_stamps):
        # Check each pulse counter is only used for a single type of pulse
        if self.receiving_pulses:
            log.error("Using a pulse counter for both pulses and offline events")
            return
        self.receiving_events = True

        self.pulse_timestamps += time_stamps
        self.count = len(self.pulse_timestamps)

    async def receive_pulse(self, di, di_value, dt_secs, counter, edge):
        """Receive active pulses on the digital input pin.

        This is **not** compatible with a PulseCounter that is already receiving events.
        """
        # Check each pulse counter is only used for a single type of pulse
        if self.receiving_events:
            log.error("Using a pulse counter for both pulses and offline events")
            return
        self.receiving_pulses = True

        if time.time() - self.start_time < self.pulse_grace_period:
            log.info(f"Ignoring pulse on di={di} with dt={dt_secs}s")
            return

        log.info(f"Received pulse on di={di} with dt={dt_secs}s")
        self.count += 1
        self.pulse_timestamps += [time.time()]
        if self.callback is not None:
            await call_maybe_async(
                self.callback, self.pin, di_value, dt_secs, self.count, edge
            )

    def receive_events(self, events):
        # Check each pulse counter is only used for a single type of pulse
        if self.receiving_pulses:
            log.error("Using a pulse counter for both pulses and offline events")
            return
        self.receiving_events = True

        for event in events:
            edge = ""
            di_value = 0
            if event.event == "DI_R":
                di_value = 1
                edge = "rising"
            elif event.event == "DI_F":
                edge = "falling"
            elif event.event == "VI":
                event = "VI"
            else:  # Could be a system event
                self.handle_system_event(event)
                continue

            timestamp = event.time / 1000 or time.time()
            dt_secs = 0
            if len(self.pulse_timestamps) > 0:
                if timestamp <= self.pulse_timestamps[-1] + 0.01:
                    log.warning(
                        f"Ignoring old event on di={event.pin} t={timestamp} latest event: {self.pulse_timestamps[-1]}"
                    )
                    continue
                dt_secs = timestamp - self.pulse_timestamps[-1]
            log.info(f"Received event on di={event.pin} with t={dt_secs}s")
            self.count += 1
            self.pulse_timestamps += [timestamp]
            if self.callback is not None:
                self.callback(self.pin, di_value, dt_secs, timestamp, self.count, edge)

    def handle_system_event(self, event: str):
        """This is called when a system event occurs. You can override this to handle system events as needed."""

    @deprecated("Use get_pulses_in_window to not damage record of pulses/events")
    def clean_pulse_timestamps(self):
        if len(self.pulse_timestamps) == 0:
            return

        ## Remove timestamps older than the rate window
        while (
            len(self.pulse_timestamps) > 0
            and self.pulse_timestamps[0] < time.time() - self.rate_window_secs
        ):
            self.pulse_timestamps.pop(0)

    def get_pulses_in_window(self):
        if len(self.pulse_timestamps) == 0:
            return []

        pulses = []
        ## Remove timestamps older than the rate window
        for timestamp in self.pulse_timestamps:
            if timestamp > self.pulse_timestamps[-1] - self.rate_window_secs:
                pulses.append(timestamp)
        return pulses

    def set_rate_window(self, rate_window_secs):
        self.rate_window_secs = rate_window_secs

    def get_rate_window(self):
        return self.rate_window_secs

    def get_pulses_per_minute(self):
        pulses = self.get_pulses_in_window()
        return len(pulses) * 60 / self.rate_window_secs

    def set_counter(self, counter):
        self.count = counter

    def get_counter(self):
        return self.count


class PlatformInterface(GRPCInterface):
    """Docker interface for interacting with the platform interface container.

    This interface allows you to interact with the platform interface gRPC service, providing access to device IO.

    Some implementations are platform-specific, and it is your responsibility to ensure that all hardware that your
    application is compatible with implements the methods you are trying to fetch. Most methods will return `None`
    if they are not supported or you pass a bad input.

    An example of bad input is requesting Digital Input #10 on a Doovit that only supports 4.
    """

    stub = platform_iface_pb2_grpc.platformIfaceStub

    def __init__(
        self, app_key: str, plt_uri: str = "localhost:50053", is_async: bool = False
    ):
        super().__init__(app_key, plt_uri, is_async)
        self.pulse_counter_listeners = []

    async def close(self):
        log.info("Closing platform interface...")
        for listener in self.pulse_counter_listeners:
            listener.cancel()

    def process_response(self, stub_call: str, response, **kwargs):
        response = super().process_response(stub_call, response, **kwargs)

        try:
            response_field = kwargs.pop("response_field")
        except KeyError:
            return response

        res = getattr(response, response_field, None)
        if isinstance(res, Iterable) and not isinstance(
            res, str
        ):  # don't iterate over strings
            res = list(res)

        if isinstance(res, list) and len(res) == 1:
            return res[0]

        return res

    def get_new_pulse_counter(
        self,
        di: int,
        edge: str = "rising",
        callback: PulseCounterCallback = None,
        rate_window_secs: int = 20,
        auto_start: bool = True,
    ) -> PulseCounter:
        """Create a new Pulse Counter for counting pulses on a digital input pin.

        Examples
        --------

        Basic pulse counter::

            def pulse_callback(di, di_value, dt_secs, count, edge):
                print(f"Pulse on di={di} with value={di_value}, dt={dt_secs}s, count={count}, edge={edge}")

            counter = self.platform_interface.get_new_pulse_counter(0, "rising", callback=pulse_callback)


        Parameters
        ----------
        di: int
            Digital input pin to listen for pulses on.
        edge: "rising" or "falling" or "both"
            The edge to listen for pulses on.
        callback : Callable
            Callback function to call when a pulse is received.
            The function should take the following arguments
            - di: int - The pin number the pulse was received on.
            - di_value: bool - The value of the digital input pin (1 for high, 0 for low).
            - dt_secs: int - The time since the last pulse in seconds.
            - count: int - The total number of pulses received.
            - edge: str - The edge that the pulse was received on ("rising" or "falling").
            The callback can be synchronous or asynchronous.

        rate_window_secs: int
            The size of window for which the rate of pulses is calculated. Default is 20.
        auto_start: bool
            Whether to automatically start listening for pulses. Default is True.
        """

        return PulseCounter(
            self,
            di,
            edge=edge,
            callback=callback,
            rate_window_secs=rate_window_secs,
            auto_start=auto_start,
        )

    def get_new_event_counter(
        self,
        di: int,
        edge: str = "rising",
        callback: PulseCounterCallback = None,
        rate_window_secs: int = 20,
        auto_collect: bool = True,
    ) -> PulseCounter:
        """Create a new Pulse Counter for counting events.

        Examples
        --------

        Basic event counter::

            counter = self.platform_interface.get_new_event_counter(0, "rising")
            print(counter.get_counter())
            print(counter.get_pulses_per_minute())


        Parameters
        ----------
        di : int
            Pin number to check events for.
        edge : "rising" or "falling" or "both"
            The edge to listen to evenets on.
        callback : PulseCounterCallback
            Callback called when an event is processed.
        rate_window_secs : int = 20
            The size of window for which the rate of events is calculated.
        auto_collect : bool = True
            Whether to automatically collect the events from the platform interface.

        Returns
        -------
        PulseCounter
            The pulse counter object for the given pin.
        """
        counter = PulseCounter(
            self,
            di,
            edge=edge,
            callback=callback,
            rate_window_secs=rate_window_secs,
            auto_start=False,
        )
        if auto_collect:
            counter.update_events()
        return counter

    def start_di_pulse_listener(
        self, di: int, callback, edge: str = "rising", start_count: int = 0
    ):
        ## Callback should be a function that takes the following arguments:
        ## di, di_value, dt_secs, counter, edge

        listener = asyncio.create_task(
            self.recv_di_pulses(di, callback, edge=edge, start_count=start_count)
        )
        self.pulse_counter_listeners.append(listener)
        listener.add_done_callback(self.pulse_counter_listeners.remove)

    async def recv_di_pulses(
        self, di: int, callback, edge: str = "rising", start_count: int = 0
    ):
        counter = start_count
        active_callbacks = set()

        while True:
            try:
                # Setup the connection to the platform interface
                async with grpc.aio.insecure_channel(self.uri) as channel:
                    channel_stream = platform_iface_pb2_grpc.platformIfaceStub(
                        channel
                    ).startPulseCounter(
                        platform_iface_pb2.pulseCounterRequest(di=di, edge=edge)
                    )

                    while True:
                        response: platform_iface_pb2.pulseCounterResponse = (
                            await channel_stream.read()
                        )
                        if response is None or response == grpc.aio.EOF:
                            log.info(f"pulseCounter for di={di} ended.")
                            break

                        log.debug(f"Received response from pulseCounter for di={di}")
                        if (
                            hasattr(response, "dt_secs")
                            and response.dt_secs is not None
                            and response.dt_secs > 0
                        ):
                            ## Increment the counter
                            counter += 1
                            ## Call the callback function with the response
                            task = await call_maybe_async(
                                callback,
                                di,
                                response.value,
                                response.dt_secs,
                                counter,
                                edge,
                                as_task=True,
                            )
                            if task:
                                active_callbacks.add(task)
                                task.add_done_callback(active_callbacks.remove)

            except asyncio.CancelledError:
                log.info(f"pulseCounter for di={di} cancelled.")
                break

            except StopAsyncIteration:
                log.info(f"pulseCounter for di={di} ended.")
                break

            except Exception as e:
                log.error(f"Error receiving pulse for di={di}: {e}", exc_info=e)
                # await asyncio.sleep(1)

            ## Loop again
            await asyncio.sleep(1)

        ## Wait for all active callbacks to finish
        while active_callbacks:
            await asyncio.wait(active_callbacks, timeout=1)

    @staticmethod
    def _cast_pins(pins):
        if isinstance(pins, int):
            return [pins]
        elif isinstance(pins, list):
            return pins
        else:
            raise ValueError(
                f"Invalid type for pins: {type(pins)}. Must be int or list."
            )
            # log.error(f"Invalid type for pins: {type(pins)}. Must be int or list."
            # return None

    @staticmethod
    def _cast_values(values):
        if isinstance(values, int):
            return [bool(values)]
        elif isinstance(values, list):
            return [bool(v) for v in values]
        else:
            raise ValueError(
                f"Invalid type for values: {type(values)}. Must be bool or list."
            )

    @staticmethod
    def _cast_ao_values(values):
        if isinstance(values, int):
            return [float(values)]
        elif isinstance(values, float):
            return [values]
        elif isinstance(values, list):
            return [float(v) for v in values]
        else:
            raise ValueError(
                f"Invalid type for values: {type(values)}. Must be float or list."
            )

    def _cast_ao_pin_values(self, pins, values):
        pins = self._cast_pins(pins)
        values = self._cast_ao_values(values)

        if len(pins) != len(values):
            if len(values) == 1:
                values = [values[0]] * len(pins)
            else:
                raise ValueError(
                    "Analogue output and value lists are not the same length."
                )

        return pins, values

    def _cast_pin_values(self, pins, values):
        pins = self._cast_pins(pins)
        values = self._cast_values(values)

        if len(pins) != len(values):
            if len(values) == 1:
                values = [values[0]] * len(pins)
            else:
                raise ValueError(
                    "Digital output and value lists are not the same length."
                )

        return pins, values

    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> str | None:
        """Test connection by sending a basic echo response to platform interface container.

        Parameters
        ----------
        message : str
            Message to send to platform interface to have echo'd as a response

        Returns
        -------
        str
            The response from platform interface.
        """
        return self.make_request(
            "TestComms",
            platform_iface_pb2.TestCommsRequest(message=message),
            response_field="response",
        )

    @cli_command()
    @maybe_async()
    def get_di(self, *di: int) -> bool | list[bool]:
        """Get digital input values.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Get a single digital input pin value::

            pin1 = await self.platform_iface.get_di(1)

        Get digital input pins 1, 2 and 3 in the same transaction::

            pin1, pin2, pin3 = await self.platform_iface.get_di(1, 2, 3)

        Parameters
        ----------
        *di
            Pin numbers to get the values of. Can be one or more integers.

        Returns
        -------
        bool | list[bool]
            Returns one or more booleans where True means the pin is high (1) and False means the pin is low (0).
            If you requested one pin, returns a single value.
            If you requested more than one pin, returns a list of values.
            Returns None if the request failed.
        """
        pins = self._cast_pins(di)
        return self.make_request(
            "getDI", platform_iface_pb2.getDIRequest(di=pins), response_field="di"
        )

    async def get_di_async(self, *di: int) -> bool | list[bool]:
        pins = self._cast_pins(di)
        return await self.make_request_async(
            "getDI", platform_iface_pb2.getDIRequest(di=pins), response_field="di"
        )

    @cli_command()
    @maybe_async()
    def get_ai(self, *ai: int) -> float | list[float]:
        """Get analogue input values.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Get a single analogue input pin value::

            pin1 = await self.platform_iface.get_ai(1)

        Get analogue input pins 1, 2 and 3 in the same transaction::

            pin1, pin2, pin3 = await self.platform_iface.get_ai(1, 2, 3)

        Parameters
        ----------
        *ai
            Pin numbers to get the values of. Can be one or more integer pin mumber.

        Returns
        -------
        float | list[float]
            If you requested one analog input, returns a single value.
            If you requested more than one pin, returns a list of values.
            Returns None if the request failed.
        """
        # Above section is to facilitate the following:
        # get_ai(1)
        # get_ai([1,4,2])

        # Proposal: get_ai(*pins)
        # allows for get_ai(1, 2, 3) or get_ai(1) or get_ai(*[1, 2, 3])

        pins = self._cast_pins(ai)
        return self.make_request(
            "getAI", platform_iface_pb2.getAIRequest(ai=pins), response_field="ai"
        )

    async def get_ai_async(self, *ai: int) -> float | list[float]:
        pins = self._cast_pins(ai)
        return await self.make_request_async(
            "getAI", platform_iface_pb2.getAIRequest(ai=pins), response_field="ai"
        )

    @cli_command()
    @maybe_async()
    def get_do(self, *do: int) -> list[bool] | None:
        """Get digital output values.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Get a single digital output pin value::

            await self.platform_iface.get_do(1)


        Get digital output pins 1, 2 and 3 in the same transaction::

            await self.platform_iface.get_do(1, 2, 3)

        Parameters
        ----------
        *do
            Pin numbers to get the values of. Can be one or more integers.

        Returns
        -------
        bool | list[bool]
            If you requested one, returns a single value.
            If you requested more than one pin, returns a list of values.
            Returns None if the request failed.
        """
        pins = self._cast_pins(do)
        return self.make_request(
            "getDO", platform_iface_pb2.getDORequest(do=pins), response_field="do"
        )

    async def get_do_async(self, *do: int) -> float | list[float]:
        pins = self._cast_pins(do)
        return await self.make_request_async(
            "getDO", platform_iface_pb2.getDORequest(do=pins), response_field="do"
        )

    @cli_command()
    @maybe_async()
    def set_do(self, do: int | list[int], value: int | list[int]) -> list[bool] | None:
        """Set digital output values.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Set a single digital output pin value::

            await self.platform_iface.set_do(1, True)

        Set digital output pin 2 to low and 3 to high in a single transaction::

            await self.platform_iface.set_do([2, 3], [False, True])

        Parameters
        ----------
        do : Union[int, list[int]]
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.
        value : Union[int, list[int]]
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.

        .. note::
            The length of the `do` and `value` lists must be the same!

        Returns
        -------
        list[bool]
            A list of digital output values that were set.
            This should ordinarily return all `True` values.
            Returns None if the request failed.
        """
        pins, values = self._cast_pin_values(do, value)
        return self.make_request(
            "setDO",
            platform_iface_pb2.setDORequest(do=pins, value=values),
            response_field="do",
        )

    async def set_do_async(self, do, value):
        pins, values = self._cast_pin_values(do, value)
        return await self.make_request_async(
            "setDO",
            platform_iface_pb2.setDORequest(do=pins, value=values),
            response_field="do",
        )

    @cli_command()
    @maybe_async()
    def schedule_do(
        self, do: int | list[int], value: bool | list[bool], in_secs: int
    ) -> None:
        """Schedule digital output values.

        This is similar to `set_do`, but schedules the change in a specified number of seconds.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Schedule a single digital output pin to be set high in 10 seconds::

            await self.platform_iface.schedule_do(1, True, 10)  # Set digital output pin 1 to high in 10 seconds


        Schedule multiple pins in 5 seconds::

            await self.platform_iface.schedule_do([2, 3], [False, True], 5)

        Parameters
        ----------
        do : Union[int, list[int]]
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.
        value : Union[bool, list[bool]]
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.
        in_secs : int
            Time in seconds to schedule the change in digital output values. Must be positive.
        """
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(
                f"Invalid value for in_secs: {in_secs}. Must be a positive integer."
            )

        pins, values = self._cast_pin_values(do, value)

        # Above section is to facilitate the following:
        # schedule_do(1, 1, 1) => [1],[1],1
        # schedule_do([1,4,2], 0, 1) => [1,4,2], [0,0,0], 1
        # schedule_do([1,4,2], [0,1,0], 1) => [1,4,2], [0,1,0], 1

        return self.make_request(
            "scheduleDO",
            platform_iface_pb2.scheduleDORequest(
                do=pins, value=values, time_secs=in_secs
            ),
            response_field="do",
        )

    async def schedule_do_async(
        self, do: int | list[int], value: bool | list[bool], in_secs: int
    ) -> None:
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(
                f"Invalid value for in_secs: {in_secs}. Must be a positive integer."
            )

        pins, values = self._cast_pin_values(do, value)
        return await self.make_request_async(
            "scheduleDO",
            platform_iface_pb2.scheduleDORequest(
                do=pins, value=values, time_secs=in_secs
            ),
            response_field="do",
        )

    @cli_command()
    @maybe_async()
    def get_ao(self, *ao: int) -> float | list[float]:
        """Get analogue output values.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Get a single analogue output pin value::

            pin1 = await self.platform_iface.get_ao(1)

        Get analogue output pins 1, 2 and 3 in the same transaction::

            await self.platform_iface.get_ao(1, 2, 3)

        Parameters
        ----------
        *ao
            Pin numbers to get the values of. Must be an integer.

        Returns
        -------
        tuple[float] | float
            If you requested multiple pins, returns a tuple of values.
            Otherwise, returns a single float.
            If the request failed, returns None.
        """
        pins = self._cast_pins(ao)
        return self.make_request(
            "getAO", platform_iface_pb2.getAORequest(ao=pins), response_field="ao"
        )

    async def get_ao_async(self, *ao: int) -> float | list[float]:
        pins = self._cast_pins(ao)
        return await self.make_request_async(
            "getAO", platform_iface_pb2.getAORequest(ao=pins), response_field="ao"
        )

    @cli_command()
    @maybe_async()
    def set_ao(
        self, ao: int | list[int], value: float | list[float]
    ) -> list[bool] | None:
        """Set analogue output values.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Set a single analogue output pin 1 to 3.3V::

            await self.platform_iface.set_ao(1, 3.3)

        Set analogue output pins 2 and 3 to 1.5V and 2.5V in a single transaction::

            await self.platform_iface.set_ao([2, 3], [1.5, 2.5])


        Parameters
        ----------
        ao : int or list[int]
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.

        value : bool or list[bool]
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.

        Returns
        -------
        list[bool]
            List of boolean values indicating whether the analogue outputs were set successfully.
        """

        # if not isinstance(value, list):
        #     value = [value]
        pins, values = self._cast_ao_pin_values(ao, value)
        return self.make_request(
            "setAO",
            platform_iface_pb2.setAORequest(ao=pins, value=values),
            response_field="ao",
        )

    async def set_ao_async(
        self, ao: int | list[int], value: float | list[float]
    ) -> list[bool]:
        pins, values = self._cast_ao_pin_values(ao, value)
        return await self.make_request_async(
            "setAO",
            platform_iface_pb2.setAORequest(ao=pins, value=values),
            response_field="ao",
        )

    @cli_command()
    @maybe_async()
    def schedule_ao(
        self, ao: int | list[int], value: bool | list[bool], in_secs: int
    ) -> None:
        """Schedule analogue output values.

        This is similar to `set_ao`, but schedules the change in a specified number of seconds.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Schedule a single analogue output pin 1 to be set to 3.3V in 10 seconds::

            await self.platform_iface.schedule_ao(1, 3.3, 10)  # Set analogue output pin 1 to 3.3V in 10 seconds


        Schedule multiple analogue output pins in 5 seconds::

            await self.platform_iface.schedule_ao([2, 3], [1.5, 2.5], 5)  # Set analogue output pins 2 and 3 to 1.5V and 2.5V respectively.


        Parameters
        ----------
        ao : int or list[int]
            Pin numbers to set the values of. Can be a single pin number or a list of pin numbers.
        value : float or list[float]
            Values to set the pins to. Can be a single value or a list of values.
            If a single value is provided, all pins will be set to that value.
        in_secs : int
            Time in seconds to schedule the change in analogue output values. Must be positive.
        """
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(
                f"Invalid value for in_secs: {in_secs}. Must be a positive integer."
            )

        pins, values = self._cast_ao_pin_values(ao, value)

        # Above section is to facilitate the following:
        # schedule_ao(1, 1, 1) => [1],[1],1
        # schedule_ao([1,4,2], 0, 1) => [1,4,2], [0,0,0], 1
        # schedule_ao([1,4,2], [0,1,0], 1) => [1,4,2], [0,1,0], 1
        return self.make_request(
            "scheduleAO",
            platform_iface_pb2.scheduleAORequest(
                ao=pins, value=values, time_secs=in_secs
            ),
            response_field="ao",
        )

    async def schedule_ao_async(
        self, ao: int | list[int], value: bool | list[bool], in_secs: int
    ) -> None:
        if not isinstance(in_secs, int) or in_secs < 0:
            raise ValueError(
                f"Invalid value for in_secs: {in_secs}. Must be a positive integer."
            )

        pins, values = self._cast_ao_pin_values(ao, value)
        return await self.make_request_async(
            "scheduleAO",
            platform_iface_pb2.scheduleAORequest(
                ao=pins, value=values, time_secs=in_secs
            ),
            response_field="ao",
        )

    @cli_command()
    @maybe_async()
    def get_system_voltage(self) -> float:
        """Get the system input voltage.

        This is the voltage supplied to the system, typically from a power supply or battery.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Get the system input voltage::

            voltage = await self.platform_iface.get_system_voltage()


        Returns
        -------
        float
            The system input voltage in volts. Returns None if the request failed.
        """
        return self.make_request(
            "getInputVoltage",
            platform_iface_pb2.getInputVoltageRequest(),
            response_field="voltage",
        )

    async def get_system_voltage_async(self) -> float:
        return await self.make_request_async(
            "getInputVoltage",
            platform_iface_pb2.getInputVoltageRequest(),
            response_field="voltage",
        )

    @cli_command()
    @maybe_async()
    def get_system_temperature(self) -> float:
        """Get the system temperature.

        On a Doovit, this is the temperature of the Raspberry Pi CM4.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Print the system temperature::

            temperature = await self.platform_iface.get_system_temperature()
            print(f"System temperature: {temperature}°C")


        Returns
        -------
        float
            The system temperature in degrees Celsius. Returns None if the request failed.
        """
        return self.make_request(
            "getTemperature",
            platform_iface_pb2.getTemperatureRequest(),
            response_field="temperature",
        )

    async def get_system_temperature_async(self):
        return await self.make_request_async(
            "getTemperature",
            platform_iface_pb2.getTemperatureRequest(),
            response_field="temperature",
        )

    @cli_command()
    @maybe_async()
    def get_location(self) -> Location:
        """Get the device location.

        Doovits with 4G cards generally implement this using the ModemManager (mmcli).

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Print the current location::

            location = await self.platform_iface.get_location()
            print(f"Latitude: {location.latitude}, Longitude: {location.longitude}, Altitude: {location.altitude}")


        Returns
        -------
        :class:`pydoover.docker.platform.Location`
            The location of the device.
            Returns None if the request failed.
        """
        return self.make_request(
            "getLocation",
            platform_iface_pb2.getLocationRequest(),
            response_field="location",
        )

    async def get_location_async(self) -> Location:
        return await self.make_request_async(
            "getLocation",
            platform_iface_pb2.getLocationRequest(),
            response_field="location",
        )

    @cli_command()
    @maybe_async()
    def reboot(self):
        """Reboot the device.

        You should **not** call this method directly, instead see
        [guide for shutting down](https://docs.doover.com/guide/app-shutdown)
        for more information on how to safely initiate a shutdown in an application.
        """
        return self.make_request("reboot", platform_iface_pb2.rebootRequest())

    async def reboot_async(self):
        # fixme: should these have async varients?
        return await self.make_request_async(
            "reboot", platform_iface_pb2.rebootRequest()
        )

    @cli_command()
    @maybe_async()
    def shutdown(self):
        """Shutdown the device.

        You should **not** call this method directly, instead see
        [guide for shutting down](https://docs.doover.com/guide/app-shutdown)
        for more information on how to safely initiate a shutdown in an application.
        """
        return self.make_request("shutdown", platform_iface_pb2.shutdownRequest())

    async def shutdown_async(self):
        # fixme: as above
        return await self.make_request_async(
            "shutdown", platform_iface_pb2.shutdownRequest()
        )

    @cli_command()
    @maybe_async()
    def get_immunity_seconds(self) -> float:
        """Get the number of seconds the device is immune for.

        Immunity is the time for which the device will ignore any shutdown requests.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Examples
        --------

        Print the number of seconds the device is immune for::

            immunity_secs = await self.platform_iface.get_immunity_seconds()
            print(f"Device immune for: {immunity_secs} seconds")


        Returns
        -------
        float
            The number of seconds the device is immune for.
        """
        return self.make_request(
            "getShutdownImmunity",
            platform_iface_pb2.getShutdownImmunityRequest(),
            response_field="immunity_secs",
        )

    async def get_immunity_seconds_async(self):
        return await self.make_request_async(
            "getShutdownImmunity",
            platform_iface_pb2.getShutdownImmunityRequest(),
            response_field="immunity_secs",
        )

    @maybe_async()
    def schedule_startup(self, time_secs: int) -> None:
        return self.make_request(
            "scheduleStartup",
            platform_iface_pb2.scheduleStartupRequest(time_secs=time_secs),
            response_field="time_secs",
        )

    async def schedule_startup_async(self, time_secs: int) -> None:
        return await self.make_request_async(
            "scheduleStartup",
            platform_iface_pb2.scheduleStartupRequest(time_secs=time_secs),
            response_field="time_secs",
        )

    @cli_command()
    @maybe_async()
    def schedule_shutdown(self, time_secs: int) -> None:
        return self.make_request(
            "scheduleShutdown",
            platform_iface_pb2.scheduleShutdownRequest(time_secs=time_secs),
            response_field="time_secs",
        )

    async def schedule_shutdown_async(self, time_secs: int) -> None:
        return await self.make_request_async(
            "scheduleShutdown",
            platform_iface_pb2.scheduleShutdownRequest(time_secs=time_secs),
            response_field="time_secs",
        )

    @cli_command()
    @maybe_async()
    def get_io_table(self):
        res = self.make_request(
            "getIoTable",
            platform_iface_pb2.getIoTableRequest(),
            response_field="io_table",
        )
        # result = json.loads("".join(self.make_request("getIoTable", platform_iface_pb2.getIoTableRequest())))
        if res is None:
            return None
        string = ""
        for i in res:
            string += i
        result = json.loads(string)
        return result

    async def get_io_table_async(self):
        res = await self.make_request_async(
            "getIoTable",
            platform_iface_pb2.getIoTableRequest(),
            response_field="io_table",
        )
        # result = json.loads("".join(await self.make_request_async("getIoTable", platform_iface_pb2.getIoTableRequest())))
        if res is None:
            return None
        string = ""
        for i in res:
            string += i
        result = json.loads(string)
        return result

    @cli_command()
    @maybe_async()
    def sync_rtc(self):
        """Synchronize the real-time clock (RTC) with the system (network) time.

        For Doovits, you shouldn't need to do this as this is handled automatically by `doovitd`.

        .. note:: This method can be used in both synchronous and asynchronous contexts.
        """
        return self.make_request("syncRtcTime", platform_iface_pb2.syncRtcTimeRequest())

    async def sync_rtc_async(self):
        return await self.make_request(
            "syncRtcTime", platform_iface_pb2.syncRtcTimeRequest()
        )

    @maybe_async()
    def get_events(self, events_from: int = 0):
        """Get all events.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Parameters
        ----------
        events_from : None or int
            Starting event id or timestamp (in milliseconds), defaults to all availible.

        Returns
        -------
        List[:class:`pydoover.docker.platform.Event`]
            List of events.
        """
        return self.make_request(
            "getEvents",
            platform_iface_pb2.getEventsRequest(events_from=events_from),
            response_field="events",
        )

    async def get_events_async(self, events_from=0):
        return await self.make_request_async(
            "getEvents",
            platform_iface_pb2.getEventsRequest(events_from=events_from),
            response_field="events",
        )

    @cli_command()
    @maybe_async()
    def get_di_events(
        self,
        di_pin: int,
        edge: str,
        include_system_events: bool = False,
        events_from: int = 0,
    ) -> list[Event]:
        """Get digital input events.

        .. note:: This method can be used in both synchronous and asynchronous contexts.

        Parameters
        ----------
        di_pin : int
            Pin number to check events for.
        edge : "rising" or "falling" or "both"
            The edge to listen to events on.
        include_system_events : bool = False
            Whether to include system events like for a doovit the cm4 turning on and off or the io board starting up.
        events_from : None or int
            Starting event id or timestamp (in milliseconds), defaults to all availible.

        Returns
        -------
        List[:class:`pydoover.docker.platform.Event`]
            List of events for the given digital input pin.
        """
        rising = False
        falling = False
        if edge == "rising":
            rising = True
        elif edge == "falling":
            falling = True
        elif edge == "both":
            rising = True
            falling = True
        return self.make_request(
            "getDIEvents",
            platform_iface_pb2.getDIEventsRequest(
                pin=di_pin,
                rising=rising,
                falling=falling,
                include_system_events=include_system_events,
                events_from=events_from,
            ),
        )

    async def get_di_events_async(
        self,
        di_pin: int,
        edge: str,
        include_system_events: bool = False,
        events_from: int = 0,
    ) -> list[Event]:
        rising = False
        falling = False
        if edge == "rising":
            rising = True
        elif edge == "falling":
            falling = True
        elif edge == "both":
            rising = True
            falling = True
        return await self.make_request_async(
            "getDIEvents",
            platform_iface_pb2.getDIEventsRequest(
                pin=di_pin,
                rising=rising,
                falling=falling,
                include_system_events=include_system_events,
                events_from=events_from,
            ),
            response_field="events",
        )


platform_iface = PlatformInterface
pulse_counter = PulseCounter



import copy
import json
import logging
import pathlib
import re

from typing import Any

log = logging.getLogger(__name__)
KEY_VALIDATOR = re.compile(r"^[ a-zA-Z0-9_-]*$")


def transform_key(key: str):
    return key.lower().replace(" ", "_")


def check_key(key: str):
    if not KEY_VALIDATOR.match(key):
        raise ValueError(
            f"Invalid config key {key}. Keys must only contain alphanumeric characters, "
            f"hyphens (-), underscores (_) and spaces ( )."
        )


class NotSet:
    pass


class Schema:
    __element_map: "dict[str, ConfigElement]"

    def add_element(self, element):
        try:
            # do this here so we don't have to override __init__
            elem_map = self.__element_map
        except AttributeError:
            # this is the first element, so create the map
            elem_map = self.__element_map = dict()

        element._name = transform_key(element.display_name)
        if element._name in elem_map:
            raise ValueError(f"Duplicate element name {element._name} not allowed.")

        elem_map[element._name] = element

    def __setattr__(self, key, value):
        if isinstance(value, ConfigElement):
            # value._name = key
            self.add_element(value)
        super().__setattr__(key, value)

    def to_dict(self):
        return {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "$id": "",
            "title": "Application Config",
            "type": "object",
            "properties": {
                name: element.to_dict()
                for name, element in self.__element_map.items()
                if isinstance(element, ConfigElement)
            },
            "additionalElements": True,
            "required": [
                name
                for name, element in self.__element_map.items()
                if isinstance(element, ConfigElement) and element.required
            ],
        }

    def _inject_deployment_config(self, config: dict[str, Any]):
        for name, value in config.items():
            try:
                elem = self.__element_map[name]
            except KeyError:
                log.info(f"Skipping unknown config key {name} ({value})")
            else:
                elem.load_data(value)

    def export(self, fp: pathlib.Path, app_name: str):
        if fp.exists():
            data = json.loads(fp.read_text())
        else:
            data = {}

        try:
            data[app_name]["config_schema"] = self.to_dict()
        except KeyError:
            data[app_name] = {"config_schema": self.to_dict()}

        fp.write_text(json.dumps(data, indent=4))


class ConfigElement:
    _type = "unknown"

    def __init__(
        self,
        display_name,
        *,
        default: Any = NotSet,
        description: str = None,
        deprecated: bool = None,
        hidden: bool = False,
    ):
        self._name = transform_key(display_name)
        self.display_name = display_name
        self.default = default
        self.description = description
        self.hidden = hidden
        self.deprecated = deprecated

        self._value = NotSet

        if (
            default is not NotSet
            and not isinstance(default, Variable)
            and default is not None
        ):
            match self._type:
                case "integer":
                    assert isinstance(default, int)
                case "number":
                    assert isinstance(default, float)
                case "string":
                    assert isinstance(default, str)
                case "boolean":
                    assert isinstance(default, bool)
                case ("array", "object"):
                    if default is not None:
                        raise ValueError(
                            "You cannot set default values for arrays and objects. It's confusing."
                        )

    @property
    def required(self):
        return self.default is NotSet

    @property
    def value(self):
        if self._value is NotSet:
            raise ValueError(f"Value for {self._name} not set. Check your config file?")
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def to_dict(self):
        payload = {
            "title": self.display_name,
            "x-name": self._name,
            "x-hidden": self.hidden,
        }

        if self._type is not None:
            payload["type"] = self._type

        if self.description is not None:
            payload["description"] = self.description

        if isinstance(self.default, Variable):
            payload["default"] = str(self.default)
        elif self.default is not NotSet:
            payload["default"] = self.default

        if self.deprecated is not None:
            payload["deprecated"] = self.deprecated

        return payload

    def load_data(self, data):
        self.value = data


class Integer(ConfigElement):
    """Represents a JSON Integer type. Internally represented as an int."""

    _type = "integer"
    value: int

    def __init__(
        self,
        display_name,
        *,
        minimum: int = None,
        exclusive_minimum: int = None,
        maximum: int = None,
        exclusive_maximum: int = None,
        multiple_of: int = None,
        **kwargs,
    ):
        super().__init__(display_name, **kwargs)
        self.minimum = minimum
        self.exclusive_minimum = exclusive_minimum
        self.maximum = maximum
        self.exclusive_maximum = exclusive_maximum
        self.multiple_of = multiple_of

    def to_dict(self):
        res = super().to_dict()
        if self.minimum is not None:
            res["minimum"] = self.minimum
        if self.exclusive_minimum is not None:
            res["exclusiveMinimum"] = self.exclusive_minimum
        if self.maximum is not None:
            res["maximum"] = self.maximum
        if self.exclusive_maximum is not None:
            res["exclusiveMaximum"] = self.exclusive_maximum
        if self.multiple_of is not None:
            res["multipleOf"] = self.multiple_of

        return res


class Number(Integer):
    """Represents a JSON Number type, for any numeric type. Internally represented as a float."""

    _type = "number"
    value: float


class Boolean(ConfigElement):
    _type = "boolean"
    value: bool


class String(ConfigElement):
    _type = "string"
    value: str

    def __init__(
        self, display_name, *, length: int = None, pattern: str = None, **kwargs
    ):
        super().__init__(display_name, **kwargs)
        self.length = length
        self.pattern = pattern

    def to_dict(self):
        res = super().to_dict()
        if self.length is not None:
            res["length"] = self.length
        if self.pattern is not None:
            res["pattern"] = self.pattern

        return res


class Enum(ConfigElement):
    _type = None

    def __init__(self, display_name, *, choices: list = None, **kwargs):
        super().__init__(display_name, **kwargs)
        self.choices = choices

        if all(isinstance(choice, str) for choice in choices):
            self._type = "string"
        elif all(isinstance(choice, float) for choice in choices):
            self._type = "number"

    def to_dict(self):
        return {
            "enum": self.choices,
            **super().to_dict(),
        }


class Array(ConfigElement):
    """Represents a JSON Array type. Internally represented as a list.

    Only a subset of JSON Schema is supported:
    - Item type
    - Minimum and maximum number of items
    - Unique items
    """

    _type = "array"

    def __init__(
        self,
        display_name,
        *,
        element: ConfigElement = None,
        min_items: int = None,
        max_items: int = None,
        unique_items: bool = None,
        **kwargs,
    ):
        if element and not isinstance(element, ConfigElement):
            raise ValueError("Many element must be a ConfigElement instance")
        if "default" in kwargs:
            raise ValueError(
                "Default value not allowed for Many elements. It's confusing."
            )

        super().__init__(display_name, **kwargs)

        self.element = element or ConfigElement("unknown")
        self.min_items = min_items
        self.max_items = max_items
        self.unique_items = unique_items

        self._elements = []

    def to_dict(self):
        res = super().to_dict()
        if self.element is not None:
            res["items"] = self.element.to_dict()
        if self.min_items is not None:
            res["minItems"] = self.min_items
        if self.max_items is not None:
            res["maxItems"] = self.max_items
        if self.unique_items is not None:
            res["uniqueItems"] = self.unique_items
        return res

    @property
    def elements(self) -> list[ConfigElement]:
        return self._elements

    def load_data(self, data):
        self._elements.clear()
        for row in data:
            elem = copy.deepcopy(self.element)
            elem.load_data(row)
            self._elements.append(elem)


class Object(ConfigElement):
    """Represents a JSON Object type."""

    _type = "object"

    def __init__(
        self,
        display_name,
        *,
        additional_elements: bool | dict[str, Any] = False,
        **kwargs,
    ):
        if "default" in kwargs:
            raise ValueError(
                "Default value not allowed for Object elements. It's confusing."
            )

        super().__init__(display_name, **kwargs)
        self._elements = {}
        self.additional_elements = additional_elements

    def __setattr__(self, key, value):
        if isinstance(value, ConfigElement):
            self.add_elements(value)
        super().__setattr__(key, value)

    def add_elements(self, *element):
        for element in element:
            if element._name in self._elements:
                raise ValueError(f"Duplicate element name {element._name} not allowed.")
            self._elements[element._name] = element

    def to_dict(self):
        res = super().to_dict()
        res["properties"] = {
            element._name: element.to_dict() for element in self._elements.values()
        }
        res["additionalElements"] = self.additional_elements
        res["required"] = [
            elem._name for elem in self._elements.values() if elem.required is True
        ]
        return res

    def load_data(self, data):
        for name, value in data.items():
            self._elements[name].load_data(value)


class Variable:
    def __init__(self, scope: str, name: str):
        self._scope = transform_key(scope)
        self._name = transform_key(name)

    def __str__(self):
        return f"${self._scope}.{self._name}"


class Application(ConfigElement):
    _type = "string"
    value: str

    def to_dict(self):
        return {"format": "doover-application", **super().to_dict()}



class DooverException(Exception):
    """General exception class for Doover API errors"""


class HTTPException(DooverException):
    """Error class for HTTP-related issues"""


class NotFound(DooverException):
    """Error raised when a resource is not found"""


class Forbidden(DooverException):
    """Error raised when access to a resource is forbidden"""



from pydoover.utils import apply_diff, generate_diff


class TestApplyDiff:
    def test_apply_diff_basic(self):
        x1 = {"a": 1, "b": 2, "c": 3}
        diff = {"c": 4}
        assert apply_diff(x1, diff) == {"a": 1, "b": 2, "c": 4}

    def test_apply_diff_remove(self):
        x1 = {"a": 1, "b": 2, "c": 3}
        diff = {"c": None}
        assert apply_diff(x1, diff) == {"a": 1, "b": 2}

    def test_apply_diff_nested(self):
        x1 = {"a": 1, "b": 2, "c": {"d": 3}}
        diff = {"c": {"d": 4}}
        assert apply_diff(x1, diff) == {"a": 1, "b": 2, "c": {"d": 4}}

    def test_apply_diff_nested_remove(self):
        x1 = {"a": 1, "b": 2, "c": {"d": 3}}
        diff = {"c": {"d": None}}
        assert apply_diff(x1, diff) == {"a": 1, "b": 2, "c": {}}

        x1 = {"a": 1, "b": 2, "c": {"d": 3, "e": 4}}
        diff = {"b": 3, "c": {"d": None}}
        assert apply_diff(x1, diff) == {"a": 1, "b": 3, "c": {"e": 4}}

    def test_apply_new_dict_old_string(self):
        x1 = "a"
        diff = {"a": 1}
        assert apply_diff(x1, diff) == {"a": 1}

    def test_apply_new_string_old_dict(self):
        x1 = {"a": 1}
        diff = "a"
        assert apply_diff(x1, diff) == "a"

    def test_apply_diff_nested_remove_no_delete(self):
        x1 = {"a": 1, "b": 2, "c": {"d": 3}}
        diff = {"c": {"d": None}}
        assert apply_diff(x1, diff, do_delete=False) == {
            "a": 1,
            "b": 2,
            "c": {"d": None},
        }

        x1 = {"a": 1, "b": 2, "c": {"d": 3, "e": 4}}
        diff = {"b": 3, "c": {"d": None}}
        assert apply_diff(x1, diff, do_delete=False) == {
            "a": 1,
            "b": 3,
            "c": {"e": 4, "d": None},
        }


class TestGenerateDiff:
    def test_generate_basic(self):
        x1 = {"a": 1, "b": 2, "c": 3}
        x2 = {"a": 1, "b": 2, "c": 4}
        assert generate_diff(x1, x2) == {"c": 4}

    def test_generate_remove(self):
        x1 = {"a": 1, "b": 2, "c": 3}
        x2 = {"a": 1, "b": 2}
        assert generate_diff(x1, x2) == {"c": None}

    def test_generate_nested(self):
        x1 = {"a": 1, "b": 2, "c": {"d": 3}}
        x2 = {"a": 1, "b": 2, "c": {"d": 4}}
        assert generate_diff(x1, x2) == {"c": {"d": 4}}

    def test_generate_nested_remove(self):
        x1 = {"a": 1, "b": 2, "c": {"d": 3}}
        x2 = {"a": 1, "b": 2, "c": {}}
        assert generate_diff(x1, x2) == {"c": {"d": None}}

        x1 = {"a": 1, "b": 2, "c": {"d": 3, "e": 4}}
        x2 = {"a": 1, "b": 2, "c": {"e": 4}}
        assert generate_diff(x1, x2) == {"c": {"d": None}}

    def test_generate_nested_same(self):
        x1 = {"a": 1, "b": 2, "c": {"d": 3}}
        x2 = {"a": 1, "b": 2, "c": {"d": 3}}
        assert generate_diff(x1, x2) == {}

    def test_generate_old_string_new_dict(self):
        x1 = "a"
        x2 = {"a": 1}
        assert generate_diff(x1, x2) == {"a": 1}

    def test_generate_new_string_old_dict(self):
        x1 = {"a": 1}
        x2 = "a"
        assert generate_diff(x1, x2) == "a"



from datetime import datetime

from typing import Union, Any

from .element import Element
from .misc import Range, Widget, NotSet


class Variable(Element):
    """Base class for UI variables. All variables should inherit from this class.

    A variable is a read-only value in the UI which is updated by a device periodically.

    Parameters
    ----------
    name: str
        The name of the variable.
    display_name: str
        The display name of the variable.
    var_type: str
        The type of the variable (e.g., "float", "string", "bool", "time").
    curr_val: Any
        The current value of the variable. If not set, defaults to NotSet.
    precision: int, optional
        The number of decimal places to round the current value to. Defaults to None.
    ranges: list[Range]
        A list of ranges associated with the variable, used for display purposes.
    earliest_data_time: datetime, optional
        The earliest time for which data is available for this variable. Defaults to None.
    """

    type = "uiVariable"

    def __init__(
        self,
        name: str,
        display_name: str,
        var_type: str,
        curr_val: Any = NotSet,
        precision: int = None,
        ranges: list[Union[Range, dict]] = None,
        earliest_data_time: datetime | None = None,
        default_zoom: str | None = None,
        **kwargs,
    ):
        # kwargs: verbose_str=verbose_str, show_activity=show_activity, form=form, graphic=graphic, layout=layout
        super().__init__(name, display_name, **kwargs)

        self.var_type = var_type
        self._curr_val = curr_val
        self.precision = precision or kwargs.pop("dec_precision", None)
        self.earliest_data_time = earliest_data_time
        self.default_zoom = default_zoom or kwargs.pop("default_zoom", None)

        self.ranges = []
        if ranges is not None:
            self.add_ranges(*ranges)

    def to_dict(self):
        result = super().to_dict()
        result["type"] = self.type
        result["varType"] = self.var_type

        curr_val = self.current_value
        if curr_val is not None:
            result["currentValue"] = curr_val

        if self.precision is not None:
            result["decPrecision"] = self.precision

        if self.earliest_data_time is not None:
            if isinstance(self.earliest_data_time, datetime):
                result["earliestDataDate"] = int(self.earliest_data_time.timestamp())
            else:
                result["earliestDataDate"] = self.earliest_data_time

        if self.default_zoom is not None:
            result["defaultZoom"] = self.default_zoom

        result["ranges"] = [r.to_dict() for r in self.ranges]
        return result

    @property
    def current_value(self):
        """Returns the current value of the variable."""
        if self._curr_val is NotSet:
            return None
        return self._curr_val

    @current_value.setter
    def current_value(self, val):
        """Updates the current value of the variable."""
        self.update(val)

    def update(self, new_value: Any) -> None:
        """Updates the current value of the variable.

        If precision is set, rounds the value to the specified number of decimal places.

        Parameters
        ----------
        new_value: Any
            The new value to set for the variable. If None, it sets the current value to None.
        """
        if self.precision is not None and new_value is not None:
            self._curr_val = round(new_value, self.precision)
        else:
            self._curr_val = new_value

    def recv_ui_state_update(self, state: dict[str, Any]) -> None:
        if self._curr_val is NotSet and "currentValue" in state:
            self.current_value = state["currentValue"]

    def add_ranges(self, *range_val: Range):
        """Adds one or more ranges to the variable.

        Parameters
        ----------
        range_val: list[Range]
            A list of Range objects to be added.
        """
        for r in range_val:
            # still support legacy dict passing of range values.
            if isinstance(r, Range):
                self.ranges.append(r)
            elif isinstance(r, dict):
                self.ranges.append(Range.from_dict(r))


class NumericVariable(Variable):
    """Represents a numeric variable in the UI, which can be an integer or a float.

    Parameters
    ----------
    name: str
        The name of the variable.
    display_name: str
        The display name of the variable.
    curr_val: Union[int, float], optional
        The current value of the variable. Defaults to None.
    precision: int, optional
        The number of decimal places to round the current value to. Defaults to None.
    ranges: list[Range], optional
        A list of ranges associated with the variable, used for display purposes. Defaults to None.
    form: Widget, optional
        A widget or string representing the form for this variable. Defaults to None.
    """

    def __init__(
        self,
        name: str,
        display_name: str,
        curr_val: Union[int, float] = None,
        precision: int = None,
        ranges: list[Range] = None,
        form: Widget | None = None,
        **kwargs,
    ):
        super().__init__(
            name,
            display_name,
            var_type="float",
            curr_val=curr_val,
            precision=precision,
            ranges=ranges,
            **kwargs,
        )
        self.form = form

    def to_dict(self):
        result = super().to_dict()
        if self.form is not None:
            result["form"] = self.form
        return result


class TextVariable(Variable):
    """Represents a text variable in the UI, which can be used to display or input string values.

    Parameters
    ----------
    name: str
        The name of the variable.
    display_name: str
        The display name of the variable.
    curr_val: str, optional
        The current value of the variable. Defaults to None.
    """

    def __init__(self, name: str, display_name: str, curr_val: str = None, **kwargs):
        # fixme: double check this type
        super().__init__(
            name, display_name, var_type="string", curr_val=curr_val, **kwargs
        )


class BooleanVariable(Variable):
    """Represents a boolean variable in the UI, which can be used to represent true/false values.

    Parameters
    ----------
    name: str
        The name of the variable.
    display_name: str
        The display name of the variable.
    curr_val: bool, optional
        The current value of the variable. Defaults to None.
    """

    def __init__(self, name: str, display_name: str, curr_val: bool = None, **kwargs):
        super().__init__(
            name, display_name, var_type="bool", curr_val=curr_val, **kwargs
        )


class DateTimeVariable(Variable):
    """Represents a date/time variable in the UI, which can be used to display or input datetime values.

    Parameters
    ----------
    name: str
        The name of the variable.
    display_name: str
        The display name of the variable.
    curr_val: Union[datetime, int], optional
        The current value of the variable, which can be a datetime object or a timestamp (int). Defaults to None.
    """

    def __init__(
        self,
        name: str,
        display_name: str,
        curr_val: Union[datetime, int] = None,
        **kwargs,
    ):
        # fixme: double check this type, and how to handle different date / time / datetime
        super().__init__(
            name, display_name, var_type="time", curr_val=curr_val, **kwargs
        )



class Object:
    """Represents a generic doover object.

    This allows you to create a "basic" object that can be passed into
    functions that expect an object, but only use the `.key` attribute.
    """

    def __init__(self, key: str):
        self.key = key

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.key == other.key
        return NotImplemented

    def to_dict(self):
        """Convert the object to a dictionary representation."""
        return {"key": self.key}



class Location:
    """Dataclass for a Location object as returned by platform interface.

    Attributes
    ----------
    latitude : float
        Latitude in degrees.
    longitude : float
        Longitude in degrees.
    altitude_m: float
        Altitude in meters above sea level.
    accuracy_m: float
        Accuracy of the location in meters.
    speed_mps: float
        Speed in meters per second.
    heading_deg: float
        Heading in degrees (0-360).
    sat_count: int
        Number of satellites used to determine the location.
    timestamp: str
        Timestamp of the location in ISO 8601 format (e.g., "2023-10-01T12:00:00Z").
    """

    latitude: float
    longitude: float
    altitude_m: float
    accuracy_m: float
    speed_mps: float
    heading_deg: float
    sat_count: int
    timestamp: str


class Event:
    """Dataclass for an Event object as returned by platform interface.

    Attributes
    ----------
    event_id : int
        Unique identifier for the event.
    event : str
        The type of event, e.g., "DI_R" for rising edge, "DI_F" for falling edge.
    pin : int
        The digital input pin number the event occurred on.
    value : str
        The value of the digital input pin at the time of the event (e.g., "1" for high, "0" for low).
    time : int
        The timestamp of the event in milliseconds since epoch.
    cm4_online : bool | None
        Whether the CM4 is online at the time of the event. This can be None if not applicable.
    """

    event_id: int
    event: str
    pin: int
    value: str
    time: int
    cm4_online: bool | None



import base64
import json
import logging
import shutil
from pathlib import Path
from typing import Any

from .object import Object

log = logging.getLogger(__name__)


class Type:
    device = "DEV"
    processor = "PRO"
    task = "TAS"
    integration = "INT"


class Visibility:
    core = "COR"
    public = "PUB"
    private = "PRI"
    internal = "INT"


class Application:
    """Represents the application configuration for a Doover application.

    This generally resides in the `doover_config.json` file in the application directory.

    Attributes:
    """

    def __init__(
        self,
        key: str,
        name: str,
        display_name: str,
        app_type: Type,
        visibility: Visibility,
        allow_many: bool,
        description: str,
        long_description: str | Path,
        depends_on: list[Object],
        owner_org: Object,
        code_repo: Object,
        repo_branch: str,
        image_name: str,
        build_args: str,
        container_registry_profile: Object,
        config_schema: dict[str, Any],
        app_base: Path,
    ):
        self.key = key

        self.name = name  # mandatory field
        self.display_name = display_name  # mandatory field

        self.type = app_type or Type.device
        self.visibility = visibility or Visibility.private

        self.allow_many = allow_many or False
        self.depends_on = depends_on or []

        self.description = description or ""

        path = app_base and long_description and app_base / long_description
        if path and path.exists():
            self.long_description = path.read_text()
        else:
            self.long_description = long_description or ""

        self.owner_org = owner_org or Object(key=None)
        self.code_repo = code_repo or Object(key=None)
        self.repo_branch = repo_branch or "main"

        self.container_registry_profile = container_registry_profile or Object(key=None)
        self.image_name = image_name
        self.build_args = build_args

        self.config_schema = config_schema or {}

        self.base_path = app_base

    @property
    def src_directory(self):
        return self.base_path / "src" / self.name.replace("-", "_")

    @classmethod
    def from_data(cls, data: dict[str, Any]) -> "Application":
        pass

    @classmethod
    def from_config(cls, data: dict, app_base: Path) -> "Application":
        return cls(
            data.get("key"),
            data.get("name"),
            data.get("display_name"),
            data.get("type"),
            data.get("visibility"),
            data.get("allow_many", False),
            data.get("description"),
            data.get("long_description"),
            [Object(key=d) for d in data.get("depends_on", [])],
            Object(key=data.get("owner_org")),
            Object(key=data.get("code_repo")),
            data.get("repo_branch"),
            data.get("image_name"),
            data.get("build_args"),
            Object(key=data.get("container_registry_profile")),
            data.get("config_schema"),
            app_base,
        )

    def to_dict(self, include_deployment_data: bool = False) -> dict[str, Any]:
        data = {
            "key": self.key,
            "name": self.name,
            "display_name": self.display_name,
            "type": self.type,
            "visibility": self.visibility,
            "allow_many": self.allow_many,
            "description": self.description,
            "long_description": self.long_description,
            "depends_on": [dep.key for dep in self.depends_on],
            "owner_org_key": self.owner_org.key,
            "code_repo_key": self.code_repo.key,
            "container_registry_profile_key": self.container_registry_profile.key,
            "repo_branch": self.repo_branch,
            "image_name": self.image_name,
            "config_schema": self.config_schema,
        }
        deployment_fp = self.base_path / "deployment"
        if include_deployment_data and deployment_fp.exists():
            tmp_fp = Path(f"/tmp/{self.name}_deployment_data.zip")
            shutil.make_archive(str(tmp_fp.with_suffix("")), "zip", deployment_fp)
            data["deployment_data"] = base64.b64encode(tmp_fp.read_bytes()).decode(
                "utf-8"
            )

        return data

    def save_to_disk(self):
        """Saves the application configuration to the disk."""
        if not self.base_path:
            raise ValueError("Application base path is not set.")

        config_path = self.base_path / "doover_config.json"
        data: dict[str, dict[str, Any]] = (
            json.loads(config_path.read_text())
            if config_path.exists()
            else {self.name: {}}
        )

        upstream = self.to_dict()
        upstream.pop("long_description", None)
        upstream.pop("owner_org_key", None)
        upstream.pop("code_repo_key", None)
        upstream.pop("container_registry_profile_key", None)

        data[self.name].update(**upstream)
        config_path.write_text(json.dumps(data, indent=4))
        log.info(f"Configuration saved to {config_path}")
